@@ -3,22 +3,26 @@ package hyperconverged
 import (
 	"context"
 	"fmt"
+	jsonpatch "github.com/evanphx/json-patch"
+	"github.com/kubevirt/hyperconverged-cluster-operator/pkg/metrics"
+	consolev1 "github.com/openshift/api/console/v1"
+	routev1 "github.com/openshift/api/route/v1"
+	rbacv1 "k8s.io/api/rbac/v1"
 	apiextensionsv1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
 	"os"
 	"reflect"
 	"strings"
 
 	"github.com/google/uuid"
-	conditionsv1 "github.com/openshift/custom-resource-status/conditions/v1"
 	operatorhandler "github.com/operator-framework/operator-lib/handler"
 	corev1 "k8s.io/api/core/v1"
 	schedulingv1 "k8s.io/api/scheduling/v1"
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
+	apimetav1 "k8s.io/apimachinery/pkg/api/meta"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/yaml"
-	"k8s.io/client-go/tools/record"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/controller"
 	"sigs.k8s.io/controller-runtime/pkg/handler"
@@ -94,7 +98,6 @@ func newReconciler(mgr manager.Manager, ci hcoutil.ClusterInfo) reconcile.Reconc
 	return &ReconcileHyperConverged{
 		client:         mgr.GetClient(),
 		scheme:         mgr.GetScheme(),
-		recorder:       mgr.GetEventRecorderFor(hcoutil.HyperConvergedName),
 		operandHandler: operands.NewOperandHandler(mgr.GetClient(), mgr.GetScheme(), ci.IsOpenshift(), hcoutil.GetEventEmitter()),
 		upgradeMode:    false,
 		ownVersion:     ownVersion,
@@ -125,6 +128,10 @@ func add(mgr manager.Manager, r reconcile.Reconciler, ci hcoutil.ClusterInfo) er
 		return err
 	}
 
+	// To limit the memory usage, the controller manager got instantiated with a custom cache
+	// that is watching only a specific set of objects with selectors.
+	// When a new object got added here, it has also to be added to the custom cache
+	// managed by getNewManagerCache()
 	secondaryResources := []client.Object{
 		&kubevirtv1.KubeVirt{},
 		&cdiv1beta1.CDI{},
@@ -132,12 +139,18 @@ func add(mgr manager.Manager, r reconcile.Reconciler, ci hcoutil.ClusterInfo) er
 		&sspv1beta1.SSP{},
 		&schedulingv1.PriorityClass{},
 		&vmimportv1beta1.VMImportConfig{},
+		&corev1.ConfigMap{},
+		&corev1.Service{},
+		&rbacv1.Role{},
+		&rbacv1.RoleBinding{},
 	}
 	if ci.IsOpenshift() {
 		secondaryResources = append(secondaryResources, []client.Object{
 			&corev1.Service{},
 			&monitoringv1.ServiceMonitor{},
 			&monitoringv1.PrometheusRule{},
+			&routev1.Route{},
+			&consolev1.ConsoleCLIDownload{},
 		}...)
 	}
 
@@ -172,7 +185,6 @@ type ReconcileHyperConverged struct {
 	// that reads objects from the cache and writes to the apiserver
 	client         client.Client
 	scheme         *runtime.Scheme
-	recorder       record.EventRecorder
 	operandHandler *operands.OperandHandler
 	upgradeMode    bool
 	ownVersion     string
@@ -223,8 +235,8 @@ func (r *ReconcileHyperConverged) Reconcile(ctx context.Context, request reconci
 		return reconcile.Result{}, err
 	}
 
-	err = r.updateHyperConverged(hcoRequest)
-	if apierrors.IsConflict(err) {
+	requeue, err := r.updateHyperConverged(hcoRequest)
+	if requeue || apierrors.IsConflict(err) {
 		result.Requeue = true
 	}
 
@@ -275,6 +287,8 @@ func (r *ReconcileHyperConverged) doReconcile(req *common.HcoRequest) (reconcile
 
 	r.setLabels(req)
 
+	updateStatusGeneration(req)
+
 	// in-memory conditions should start off empty. It will only ever hold
 	// negative conditions (!Available, Degraded, Progressing)
 	req.Conditions = common.NewHcoConditions()
@@ -346,6 +360,13 @@ func (r *ReconcileHyperConverged) EnsureOperandAndComplete(req *common.HcoReques
 	return reconcile.Result{}, nil
 }
 
+func updateStatusGeneration(req *common.HcoRequest) {
+	if req.Instance.ObjectMeta.Generation != req.Instance.Status.ObservedGeneration {
+		req.Instance.Status.ObservedGeneration = req.Instance.ObjectMeta.Generation
+		req.StatusDirty = true
+	}
+}
+
 // getHyperConverged gets the HyperConverged resource from the Kubernetes API.
 func (r *ReconcileHyperConverged) getHyperConverged(req *common.HcoRequest) (*hcov1beta1.HyperConverged, error) {
 	instance := &hcov1beta1.HyperConverged{}
@@ -371,7 +392,7 @@ func (r *ReconcileHyperConverged) getHyperConverged(req *common.HcoRequest) (*hc
 }
 
 // updateHyperConverged updates the HyperConverged resource according to its state in the request.
-func (r *ReconcileHyperConverged) updateHyperConverged(request *common.HcoRequest) error {
+func (r *ReconcileHyperConverged) updateHyperConverged(request *common.HcoRequest) (bool, error) {
 
 	// Since the status subresource is enabled for the HyperConverged kind,
 	// we need to update the status and the metadata separately.
@@ -379,32 +400,22 @@ func (r *ReconcileHyperConverged) updateHyperConverged(request *common.HcoReques
 	// In addition, spec changes are removed by status update, but since status update done first, we need
 	// to store the spec and recover it after status update
 
-	var spec hcov1beta1.HyperConvergedSpec
 	if request.Dirty {
-		request.Instance.Spec.DeepCopyInto(&spec)
+		err := r.updateHyperConvergedSpecMetadata(request)
+		if err != nil {
+			r.logHyperConvergedUpdateError(request, err, "Failed to update HCO CR")
+			return false, err
+		}
+		return true, nil
 	}
 
 	err := r.updateHyperConvergedStatus(request)
 	if err != nil {
 		r.logHyperConvergedUpdateError(request, err, "Failed to update HCO Status")
-		return err
+		return false, err
 	}
 
-	// restore spec
-	if request.Dirty {
-		request.Instance.Spec = spec
-	}
-
-	// Doing it here because status.update overrides spec for some reason
-	r.recoverHCOVersion(request)
-
-	err = r.updateHyperConvergedSpecMetadata(request)
-	if err != nil {
-		r.logHyperConvergedUpdateError(request, err, "Failed to update HCO CR")
-		return err
-	}
-
-	return nil
+	return false, nil
 }
 
 // updateHyperConvergedSpecMetadata updates the HyperConverged resource's spec and metadata.
@@ -446,11 +457,12 @@ func (r *ReconcileHyperConverged) validateNamespace(req *common.HcoRequest) (boo
 	// Ignore invalid requests
 	if req.NamespacedName != hco {
 		req.Logger.Info("Invalid request", "HyperConverged.Namespace", hco.Namespace, "HyperConverged.Name", hco.Name)
-		req.Conditions.SetStatusCondition(conditionsv1.Condition{
-			Type:    hcov1beta1.ConditionReconcileComplete,
-			Status:  corev1.ConditionFalse,
-			Reason:  invalidRequestReason,
-			Message: fmt.Sprintf(invalidRequestMessageFormat, hco.Name, hco.Namespace),
+		req.Conditions.SetStatusCondition(metav1.Condition{
+			Type:               hcov1beta1.ConditionReconcileComplete,
+			Status:             metav1.ConditionFalse,
+			Reason:             invalidRequestReason,
+			Message:            fmt.Sprintf(invalidRequestMessageFormat, hco.Name, hco.Namespace),
+			ObservedGeneration: req.Instance.Generation,
 		})
 		r.updateConditions(req)
 		return false, nil
@@ -460,38 +472,41 @@ func (r *ReconcileHyperConverged) validateNamespace(req *common.HcoRequest) (boo
 
 func (r *ReconcileHyperConverged) setInitialConditions(req *common.HcoRequest) {
 	req.Instance.Status.UpdateVersion(hcoVersionName, r.ownVersion)
-	req.Instance.Spec.Version = r.ownVersion
-	req.Dirty = true
 
-	req.Conditions.SetStatusCondition(conditionsv1.Condition{
-		Type:    hcov1beta1.ConditionReconcileComplete,
-		Status:  corev1.ConditionUnknown, // we just started trying to reconcile
-		Reason:  reconcileInit,
-		Message: reconcileInitMessage,
+	req.Conditions.SetStatusCondition(metav1.Condition{
+		Type:               hcov1beta1.ConditionReconcileComplete,
+		Status:             metav1.ConditionUnknown, // we just started trying to reconcile
+		Reason:             reconcileInit,
+		Message:            reconcileInitMessage,
+		ObservedGeneration: req.Instance.Generation,
 	})
-	req.Conditions.SetStatusCondition(conditionsv1.Condition{
-		Type:    conditionsv1.ConditionAvailable,
-		Status:  corev1.ConditionFalse,
-		Reason:  reconcileInit,
-		Message: reconcileInitMessage,
+	req.Conditions.SetStatusCondition(metav1.Condition{
+		Type:               hcov1beta1.ConditionAvailable,
+		Status:             metav1.ConditionFalse,
+		Reason:             reconcileInit,
+		Message:            reconcileInitMessage,
+		ObservedGeneration: req.Instance.Generation,
 	})
-	req.Conditions.SetStatusCondition(conditionsv1.Condition{
-		Type:    conditionsv1.ConditionProgressing,
-		Status:  corev1.ConditionTrue,
-		Reason:  reconcileInit,
-		Message: reconcileInitMessage,
+	req.Conditions.SetStatusCondition(metav1.Condition{
+		Type:               hcov1beta1.ConditionProgressing,
+		Status:             metav1.ConditionTrue,
+		Reason:             reconcileInit,
+		Message:            reconcileInitMessage,
+		ObservedGeneration: req.Instance.Generation,
 	})
-	req.Conditions.SetStatusCondition(conditionsv1.Condition{
-		Type:    conditionsv1.ConditionDegraded,
-		Status:  corev1.ConditionFalse,
-		Reason:  reconcileInit,
-		Message: reconcileInitMessage,
+	req.Conditions.SetStatusCondition(metav1.Condition{
+		Type:               hcov1beta1.ConditionDegraded,
+		Status:             metav1.ConditionFalse,
+		Reason:             reconcileInit,
+		Message:            reconcileInitMessage,
+		ObservedGeneration: req.Instance.Generation,
 	})
-	req.Conditions.SetStatusCondition(conditionsv1.Condition{
-		Type:    conditionsv1.ConditionUpgradeable,
-		Status:  corev1.ConditionUnknown,
-		Reason:  reconcileInit,
-		Message: reconcileInitMessage,
+	req.Conditions.SetStatusCondition(metav1.Condition{
+		Type:               hcov1beta1.ConditionUpgradeable,
+		Status:             metav1.ConditionUnknown,
+		Reason:             reconcileInit,
+		Message:            reconcileInitMessage,
+		ObservedGeneration: req.Instance.Generation,
 	})
 
 	r.updateConditions(req)
@@ -631,83 +646,93 @@ func (r *ReconcileHyperConverged) aggregateComponentConditions(req *common.HcoRe
 		    you will only see CDI as it updates last).
 	*/
 	allComponentsAreUp := req.Conditions.IsEmpty()
-	req.Conditions.SetStatusCondition(conditionsv1.Condition{
-		Type:    hcov1beta1.ConditionReconcileComplete,
-		Status:  corev1.ConditionTrue,
-		Reason:  reconcileCompleted,
-		Message: reconcileCompletedMessage,
+	req.Conditions.SetStatusCondition(metav1.Condition{
+		Type:               hcov1beta1.ConditionReconcileComplete,
+		Status:             metav1.ConditionTrue,
+		Reason:             reconcileCompleted,
+		Message:            reconcileCompletedMessage,
+		ObservedGeneration: req.Instance.Generation,
 	})
 
-	if req.Conditions.HasCondition(conditionsv1.ConditionDegraded) { // (#chart 1)
+	if req.Conditions.HasCondition(hcov1beta1.ConditionDegraded) { // (#chart 1)
 
-		req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 2,3)
-			Type:    conditionsv1.ConditionProgressing,
-			Status:  corev1.ConditionFalse,
-			Reason:  reconcileCompleted,
-			Message: reconcileCompletedMessage,
+		req.Conditions.SetStatusConditionIfUnset(metav1.Condition{ // (#chart 2,3)
+			Type:               hcov1beta1.ConditionProgressing,
+			Status:             metav1.ConditionFalse,
+			Reason:             reconcileCompleted,
+			Message:            reconcileCompletedMessage,
+			ObservedGeneration: req.Instance.Generation,
 		})
 
-		req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 4,5)
-			Type:    conditionsv1.ConditionUpgradeable,
-			Status:  corev1.ConditionFalse,
-			Reason:  commonDegradedReason,
-			Message: "HCO is not Upgradeable due to degraded components",
+		req.Conditions.SetStatusConditionIfUnset(metav1.Condition{ // (#chart 4,5)
+			Type:               hcov1beta1.ConditionUpgradeable,
+			Status:             metav1.ConditionFalse,
+			Reason:             commonDegradedReason,
+			Message:            "HCO is not Upgradeable due to degraded components",
+			ObservedGeneration: req.Instance.Generation,
 		})
 
-		req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 6,7)
-			Type:    conditionsv1.ConditionAvailable,
-			Status:  corev1.ConditionFalse,
-			Reason:  commonDegradedReason,
-			Message: "HCO is not available due to degraded components",
+		req.Conditions.SetStatusConditionIfUnset(metav1.Condition{ // (#chart 6,7)
+			Type:               hcov1beta1.ConditionAvailable,
+			Status:             metav1.ConditionFalse,
+			Reason:             commonDegradedReason,
+			Message:            "HCO is not available due to degraded components",
+			ObservedGeneration: req.Instance.Generation,
 		})
 
 	} else {
 
 		// Degraded is not found. add it.
-		req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 8)
-			Type:    conditionsv1.ConditionDegraded,
-			Status:  corev1.ConditionFalse,
-			Reason:  reconcileCompleted,
-			Message: reconcileCompletedMessage,
+		req.Conditions.SetStatusCondition(metav1.Condition{ // (#chart 8)
+			Type:               hcov1beta1.ConditionDegraded,
+			Status:             metav1.ConditionFalse,
+			Reason:             reconcileCompleted,
+			Message:            reconcileCompletedMessage,
+			ObservedGeneration: req.Instance.Generation,
 		})
 
-		if req.Conditions.HasCondition(conditionsv1.ConditionProgressing) { // (#chart 9)
+		if req.Conditions.HasCondition(hcov1beta1.ConditionProgressing) { // (#chart 9)
 
-			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 10,11)
-				Type:    conditionsv1.ConditionUpgradeable,
-				Status:  corev1.ConditionFalse,
-				Reason:  commonProgressingReason,
-				Message: "HCO is not Upgradeable due to progressing components",
+			req.Conditions.SetStatusConditionIfUnset(metav1.Condition{ // (#chart 10,11)
+				Type:               hcov1beta1.ConditionUpgradeable,
+				Status:             metav1.ConditionFalse,
+				Reason:             commonProgressingReason,
+				Message:            "HCO is not Upgradeable due to progressing components",
+				ObservedGeneration: req.Instance.Generation,
 			})
 
-			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 12,13)
-				Type:    conditionsv1.ConditionAvailable,
-				Status:  corev1.ConditionTrue,
-				Reason:  reconcileCompleted,
-				Message: reconcileCompletedMessage,
+			req.Conditions.SetStatusConditionIfUnset(metav1.Condition{ // (#chart 12,13)
+				Type:               hcov1beta1.ConditionAvailable,
+				Status:             metav1.ConditionTrue,
+				Reason:             reconcileCompleted,
+				Message:            reconcileCompletedMessage,
+				ObservedGeneration: req.Instance.Generation,
 			})
 
 		} else {
 
-			req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 14)
-				Type:    conditionsv1.ConditionProgressing,
-				Status:  corev1.ConditionFalse,
-				Reason:  reconcileCompleted,
-				Message: reconcileCompletedMessage,
+			req.Conditions.SetStatusCondition(metav1.Condition{ // (#chart 14)
+				Type:               hcov1beta1.ConditionProgressing,
+				Status:             metav1.ConditionFalse,
+				Reason:             reconcileCompleted,
+				Message:            reconcileCompletedMessage,
+				ObservedGeneration: req.Instance.Generation,
 			})
 
-			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 15,16)
-				Type:    conditionsv1.ConditionUpgradeable,
-				Status:  corev1.ConditionTrue,
-				Reason:  reconcileCompleted,
-				Message: reconcileCompletedMessage,
+			req.Conditions.SetStatusConditionIfUnset(metav1.Condition{ // (#chart 15,16)
+				Type:               hcov1beta1.ConditionUpgradeable,
+				Status:             metav1.ConditionTrue,
+				Reason:             reconcileCompleted,
+				Message:            reconcileCompletedMessage,
+				ObservedGeneration: req.Instance.Generation,
 			})
 
-			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 17,18)
-				Type:    conditionsv1.ConditionAvailable,
-				Status:  corev1.ConditionTrue,
-				Reason:  reconcileCompleted,
-				Message: reconcileCompletedMessage,
+			req.Conditions.SetStatusConditionIfUnset(metav1.Condition{ // (#chart 17,18)
+				Type:               hcov1beta1.ConditionAvailable,
+				Status:             metav1.ConditionTrue,
+				Reason:             reconcileCompleted,
+				Message:            reconcileCompletedMessage,
+				ObservedGeneration: req.Instance.Generation,
 			})
 
 		}
@@ -724,15 +749,12 @@ func (r *ReconcileHyperConverged) completeReconciliation(req *common.HcoRequest)
 	if allComponentsAreUp {
 		req.Logger.Info("No component operator reported negatively")
 
-		// if in upgrade mode, and all the components are upgraded - upgrade is completed
-		if r.upgradeMode && req.ComponentUpgradeInProgress {
+		// if in upgrade mode, and all the components are upgraded, and nothing pending to be written - upgrade is completed
+		if r.upgradeMode && req.ComponentUpgradeInProgress && !req.Dirty {
 			// update the new version only when upgrade is completed
 			req.Instance.Status.UpdateVersion(hcoVersionName, r.ownVersion)
 			req.StatusDirty = true
 
-			req.Instance.Spec.Version = r.ownVersion
-			req.Dirty = true
-
 			r.upgradeMode = false
 			req.ComponentUpgradeInProgress = false
 			req.Logger.Info(fmt.Sprintf("Successfuly upgraded to version %s", r.ownVersion))
@@ -746,11 +768,12 @@ func (r *ReconcileHyperConverged) completeReconciliation(req *common.HcoRequest)
 
 	if r.upgradeMode {
 		// override the Progressing condition during upgrade
-		req.Conditions.SetStatusCondition(conditionsv1.Condition{
-			Type:    conditionsv1.ConditionProgressing,
-			Status:  corev1.ConditionTrue,
-			Reason:  "HCOUpgrading",
-			Message: "HCO is now upgrading to version " + r.ownVersion,
+		req.Conditions.SetStatusCondition(metav1.Condition{
+			Type:               hcov1beta1.ConditionProgressing,
+			Status:             metav1.ConditionTrue,
+			Reason:             "HCOUpgrading",
+			Message:            "HCO is now upgrading to version " + r.ownVersion,
+			ObservedGeneration: req.Instance.Generation,
 		})
 	}
 
@@ -773,22 +796,31 @@ func (r *ReconcileHyperConverged) completeReconciliation(req *common.HcoRequest)
 
 // This function is used to exit from the reconcile function, updating the conditions and returns the reconcile result
 func (r *ReconcileHyperConverged) updateConditions(req *common.HcoRequest) {
+	conditions := make([]metav1.Condition, len(req.Instance.Status.Conditions))
+	copy(conditions, req.Instance.Status.Conditions)
+
 	for _, condType := range common.HcoConditionTypes {
 		cond, found := req.Conditions[condType]
 		if !found {
-			cond = conditionsv1.Condition{
-				Type:    condType,
-				Status:  corev1.ConditionUnknown,
-				Message: "Unknown Status",
+			cond = metav1.Condition{
+				Type:               condType,
+				Status:             metav1.ConditionUnknown,
+				Message:            "Unknown Status",
+				Reason:             "StatusUnknown",
+				ObservedGeneration: req.Instance.ObjectMeta.Generation,
 			}
 		}
-		conditionsv1.SetStatusCondition(&req.Instance.Status.Conditions, cond)
+
+		apimetav1.SetStatusCondition(&conditions, cond)
 	}
 
 	// Detect a "TaintedConfiguration" state, and raise a corresponding event
-	r.detectTaintedConfiguration(req)
+	r.detectTaintedConfiguration(req, &conditions)
 
-	req.StatusDirty = true
+	if !reflect.DeepEqual(conditions, req.Instance.Status.Conditions) {
+		req.Instance.Status.Conditions = conditions
+		req.StatusDirty = true
+	}
 }
 
 func (r *ReconcileHyperConverged) setLabels(req *common.HcoRequest) {
@@ -801,47 +833,56 @@ func (r *ReconcileHyperConverged) setLabels(req *common.HcoRequest) {
 	}
 }
 
-func (r *ReconcileHyperConverged) detectTaintedConfiguration(req *common.HcoRequest) {
-	conditionExists := conditionsv1.IsStatusConditionTrue(req.Instance.Status.Conditions,
-		hcov1beta1.ConditionTaintedConfiguration)
+func (r *ReconcileHyperConverged) detectTaintedConfiguration(req *common.HcoRequest, conditions *[]metav1.Condition) {
+	conditionExists := apimetav1.IsStatusConditionTrue(req.Instance.Status.Conditions, hcov1beta1.ConditionTaintedConfiguration)
 
 	// A tainted configuration state is indicated by the
 	// presence of at least one of the JSON Patch annotations
 	tainted := false
 	for _, jpa := range JSONPatchAnnotationNames {
-		_, exists := req.Instance.ObjectMeta.Annotations[jpa]
+		NumOfChanges := 0
+		jsonPatch, exists := req.Instance.ObjectMeta.Annotations[jpa]
 		if exists {
-			tainted = true
-			break
+			if NumOfChanges = getNumOfChangesJsonPatch(jsonPatch); NumOfChanges > 0 {
+				tainted = true
+			}
 		}
+		metrics.HcoMetrics.SetUnsafeModificationCount(NumOfChanges, jpa)
 	}
 
 	if tainted {
-		conditionsv1.SetStatusCondition(&req.Instance.Status.Conditions, conditionsv1.Condition{
-			Type:    hcov1beta1.ConditionTaintedConfiguration,
-			Status:  corev1.ConditionTrue,
-			Reason:  taintedConfigurationReason,
-			Message: taintedConfigurationMessage,
+		apimetav1.SetStatusCondition(conditions, metav1.Condition{
+			Type:               hcov1beta1.ConditionTaintedConfiguration,
+			Status:             metav1.ConditionTrue,
+			Reason:             taintedConfigurationReason,
+			Message:            taintedConfigurationMessage,
+			ObservedGeneration: req.Instance.ObjectMeta.Generation,
 		})
 
 		if !conditionExists {
 			// Only log at "first occurrence" of detection
 			req.Logger.Info("Detected tainted configuration state for HCO")
-			req.StatusDirty = true
 		}
 	} else { // !tainted
 
 		// For the sake of keeping the JSONPatch backdoor in low profile,
 		// we just remove the condition instead of False'ing it.
 		if conditionExists {
-			conditionsv1.RemoveStatusCondition(&req.Instance.Status.Conditions, hcov1beta1.ConditionTaintedConfiguration)
+			apimetav1.RemoveStatusCondition(conditions, hcov1beta1.ConditionTaintedConfiguration)
 
 			req.Logger.Info("Detected untainted configuration state for HCO")
-			req.StatusDirty = true
 		}
 	}
 }
 
+func getNumOfChangesJsonPatch(jsonPatch string) int {
+	patches, err := jsonpatch.DecodePatch([]byte(jsonPatch))
+	if err != nil {
+		return 0
+	}
+	return len(patches)
+}
+
 func (r *ReconcileHyperConverged) firstLoopInitialization(request *common.HcoRequest) {
 	// Reload eventEmitter.
 	// The client should now find all the required resources.
@@ -854,20 +895,6 @@ func (r *ReconcileHyperConverged) firstLoopInitialization(request *common.HcoReq
 	r.firstLoop = false
 }
 
-// recoverHCOVersion recovers Spec.Version if upgrade missed when upgrade completed
-func (r *ReconcileHyperConverged) recoverHCOVersion(request *common.HcoRequest) {
-	knownHcoVersion, versionFound := request.Instance.Status.GetVersion(hcoVersionName)
-
-	if !r.upgradeMode &&
-		versionFound &&
-		(knownHcoVersion == r.ownVersion) &&
-		(request.Instance.Spec.Version != r.ownVersion) {
-
-		request.Instance.Spec.Version = r.ownVersion
-		request.Dirty = true
-	}
-}
-
 // This function performs migrations before starting the upgrade process
 // return true if the HyperConverged CR was modified; else, return false
 //
@@ -938,7 +965,12 @@ func (r *ReconcileHyperConverged) migrateBeforeUpgrade(req *common.HcoRequest) (
 		return false, err
 	}
 
-	return kvConfigModified || cdiConfigModified || imsConfigModified, nil
+	defaultsAmended, err := r.amendBadDefaults(req)
+	if err != nil {
+		return false, err
+	}
+
+	return kvConfigModified || cdiConfigModified || imsConfigModified || defaultsAmended, nil
 }
 
 func (r ReconcileHyperConverged) migrateKvConfigurations(req *common.HcoRequest) (bool, error) {
@@ -1007,7 +1039,17 @@ func (r ReconcileHyperConverged) migrateImsConfigurations(req *common.HcoRequest
 			modified = true
 		}
 	}
+	return modified, nil
+}
 
+func (r ReconcileHyperConverged) amendBadDefaults(req *common.HcoRequest) (bool, error) {
+	modified := false
+	if req.Instance.Spec.LiveMigrationConfig.BandwidthPerMigration != nil && *req.Instance.Spec.LiveMigrationConfig.BandwidthPerMigration == "64Mi" {
+		req.Logger.Info("rejecting spec.livemigrationconfig.bandwidthpermigration==64Mi because it was a bad default value")
+		req.Instance.Spec.LiveMigrationConfig.BandwidthPerMigration = nil
+		modified = true
+		req.Dirty = true
+	}
 	return modified, nil
 }
 
